@@ -2,11 +2,11 @@
 
 ;; Copyright (C) 2020  Free Software Foundation, Inc.
 
-;; Author: Philip K. <philipk@posteo.net>
-;; Version: 2.2.0
+;; Author: Philip Kaludercic <philipk@posteo.net>
+;; Version: 2.2.1
 ;; Keywords: unix, processes, convenience
 ;; Package-Requires: ((emacs "24.1"))
-;; URL: https://git.sr.ht/~zge/bang
+;; URL: https://git.sr.ht/~pkal/shell-command-plus
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -61,11 +61,13 @@
 
 (eval-when-compile (require 'rx))
 (eval-when-compile (require 'pcase))
+(require 'diff)
+(require 'info)
 
 ;;; Code:
 
 (defgroup shell-command+ nil
-  "An extended `shell-command'"
+  "An extended `shell-command'."
   :group 'external
   :prefix "shell-command+-")
 
@@ -88,6 +90,10 @@ handlers if the symbol (eg. `man') is contained in the list."
   "Flip the meaning of < and > at the beginning of a command."
   :type 'boolean)
 
+(defcustom shell-command+-enable-file-substitution t
+  "Enable the substitution of \"%s\" with the current file name."
+  :type 'boolean)
+
 (defcustom shell-command+-substitute-alist
   (cond ((eq shell-command+-use-eshell t)
          (require 'eshell)
@@ -134,21 +140,35 @@ the command string"
 
 
 
+(defconst shell-command+-token-regexp
+  (rx bos (* space)
+      (or (: ?\"
+             (group-n 1 (* (or (: ?\\ anychar) (not (any ?\\ ?\")))))
+             ?\")
+          (: ?\'
+             (group-n 1 (* (or (: ?\\ anychar) (not (any ?\\ ?\')))))
+             ?\')
+          (group (+ (not (any space ?\\ ?\" ?\')))
+                 (* ?\\ anychar (* (not (any space ?\\ ?\" ?\')))))))
+  "Regular expression for tokenizing shell commands.")
+
 (defun shell-command+-tokenize (command &optional expand)
   "Return list of tokens of COMMAND.
 If EXPAND is non-nil, expand wildcards."
   (let ((pos 0) tokens)
-    (while (string-match
-            (rx bos (* space)
-                (or (: ?\" (group (* (not ?\"))) ?\")
-                    (: (group (+ (not (any ?\" space)))))))
-            (substring command pos))
+    (while (string-match shell-command+-token-regexp (substring command pos))
       (push (let ((tok (match-string 2 (substring command pos))))
               (if (and expand tok)
                   (or (file-expand-wildcards tok) (list tok))
-                (list (or (match-string 2 (substring command pos))
-                          (match-string 1 (substring command pos))))))
+                (list (replace-regexp-in-string
+                       (rx (* ?\\ ?\\) (group ?\\ (group anychar)))
+                       "\\2"
+                       (or (match-string 2 (substring command pos))
+                           (match-string 1 (substring command pos)))
+                       nil nil 1))))
             tokens)
+      (when (= (match-end 0) 0)
+        (error "Zero-width token parsed"))
       (setq pos (+ pos (match-end 0))))
     (unless (= pos (length command))
       (error "Tokenization error at %s" (substring command pos)))
@@ -156,7 +176,13 @@ If EXPAND is non-nil, expand wildcards."
 
 (defun shell-command+-cmd-grep (command)
   "Convert COMMAND into a `grep' call."
-  (grep (mapconcat #'identity (shell-command+-tokenize command t) " ")))
+  (grep-compute-defaults)
+  (pcase-let ((`(,cmd . ,args) (shell-command+-tokenize command t)))
+    (grep (mapconcat #'identity
+                     (cons (replace-regexp-in-string
+                            (concat "\\`" grep-program) cmd grep-command)
+                           args)
+                     " "))))
 
 (defun shell-command+-cmd-find (command)
   "Convert COMMAND into a `find-dired' call."
@@ -258,11 +284,13 @@ proper upwards directory pointers.  This means that '....' becomes
                  'literal))
           (match-string-no-properties 4 command)
           (condition-case nil
-              (replace-regexp-in-string
-               (rx (* ?\\ ?\\) (or ?\\ (group "%")))
-               buffer-file-name
-               (match-string-no-properties 3 command)
-               nil nil 1)
+              (if shell-command+-enable-file-substitution
+                  (replace-regexp-in-string
+                   (rx (* ?\\ ?\\) (or ?\\ (group "%")))
+                   buffer-file-name
+                   (match-string-no-properties 3 command)
+                   nil nil 1)
+                (match-string-no-properties 3 command))
             (error (match-string-no-properties 3 command))))))
 
 ;;;###autoload
@@ -291,7 +319,11 @@ These extentions can all be combined with one-another.
 
 In case a region is active, `shell-command+' will only work with the region
 between BEG and END.  Otherwise the whole buffer is processed."
-  (interactive (list (read-shell-command shell-command+-prompt)
+  (interactive (list (read-shell-command
+                      (if shell-command-prompt-show-cwd
+                          (format shell-command+-prompt
+                                  (abbreviate-file-name default-directory))
+                        shell-command+-prompt))
                      (if (use-region-p) (region-beginning) (point-min))
                      (if (use-region-p) (region-end) (point-max))))
   (pcase-let* ((`(,path ,mode ,command ,rest) (shell-command+-parse command))
